/*
   Copyright 2007-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.gui;

import java.awt.Component;

import javax.swing.*;

import org.apache.log4j.Logger;

import com.notuvy.util.ThreadWorker;
import com.notuvy.util.WallClock;

/**
 * An abstract dialog that is displayed only for a limited amount of time.
 *
 * @author  murali
 */
public abstract class TimedDialog implements Runnable {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger(TimedDialog.class);

    private static final ThreadWorker THREAD_CLOSER = new ThreadWorker("TimedDialog");


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private JDialog vDialog;
    private final Component vParent;
    private final String vTitle;
    private final String vMessage;
    private final int vTimeout;
    private long vEndTime;
    private boolean vClosed = false;
    private boolean fExpired = false;
    private boolean fInitialized = false;


    //------------------------------------------------------------
    //- Constructors

    protected TimedDialog(Component pParent, String pTitle, String pMessage, int pTimeout) {
        vParent = pParent;
        vTitle = pTitle;
        vMessage = pMessage;
        vTimeout = pTimeout;
    }


    //------------------------------------------------------------
    //- Accessors

    private JDialog getDialog() {
        return vDialog;
    }

    protected Component getParent() {
        return vParent;
    }

    protected String getTitle() {
        return vTitle;
    }

    protected String getMessage() {
        return vMessage;
    }

    private int getTimeout() {
        return vTimeout;
    }

    private long getEndTime() {
        return vEndTime;
    }

    public boolean getClosed() {
        return vClosed;
    }

    public boolean isInitialized() {
        return fInitialized;
    }

    public boolean isExpired() {
        return fExpired;
    }


    //------------------------------------------------------------
    //- Settors

    private void setDialog(JDialog pDialog) {
        vDialog = pDialog;
    }

    private void setEndTime(long pEndTime) {
        vEndTime = pEndTime;
    }

    public void setClosed(boolean pClosed) {
        vClosed = pClosed;
    }

    private void setExpired(boolean pExpired) {
        fExpired = pExpired;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions


    //------------------------------------------------------------
    //- Public Interface Functions

    public void show() {
        JOptionPane pane = createOptionPane();
        setDialog(pane.createDialog(getParent(), getTitle()));

        setEndTime(WallClock.currentTimeMillis() + getTimeout());
        THREAD_CLOSER.enqueueAsync(this);

        fInitialized = true;
        getDialog().setVisible(true);  // This will display the UI window and suspend this thread.

        setClosed(true);
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    protected abstract JOptionPane createOptionPane();


    public void run() {
        do {
            try {
                Thread.sleep(200);
            } catch (InterruptedException ie) {
                // nothing
            }
            setExpired(WallClock.currentTimeMillis() >= getEndTime());
        } while (!getClosed() && !isExpired());

        if (!getClosed()) {
            getDialog().dispose();
        }
    }


    //------------------------------------------------------------
    //- Inner Classes



    //------------------------------------------------------------
    //- Main



}
