/*
   Copyright 2006-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.notuvy.net;

import java.io.*;
import java.awt.*;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.GetMethod;

import org.apache.log4j.Logger;

import com.notuvy.cmd.Command;
import com.notuvy.cmd.CommandArgs;
import com.notuvy.cmd.CommandAction;
import com.notuvy.util.Progressable;
import com.notuvy.util.DummyProgress;
import com.notuvy.util.ThreadMonitor;

/**
 * A very simple utility for downloading a resource from an URL.
 *
 * @author  murali
 */
public class Download implements Runnable {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger(Download.class);

    private static final int BUFFER_SIZE = 1024;
    private static final int SOCKET_READ_TIMEOUT = 60000;  // one minute.


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private final String vUrl;
    private final File vFile;
    private Progressable vProgress = new DummyProgress();
    private boolean vIsAborted = false;
    private boolean vIsSuccessful = false;


    //------------------------------------------------------------
    //- Constructors

    public Download(String pUrl) {
        this (pUrl, null);
    }

    public Download(String pUrl, File pFile) {
        vUrl = pUrl;
        vFile = pFile;
    }


    //------------------------------------------------------------
    //- Accessors

    public String getUrl() {
        return vUrl;
    }

    public File getFile() {
        return vFile;
    }

    public Progressable getProgress() {
        return vProgress;
    }

    public boolean getIsAborted() {
        return vIsAborted;
    }

    public boolean getIsSuccessful() {
        return vIsSuccessful;
    }


    //------------------------------------------------------------
    //- Settors

    public void setProgress(Progressable pProgress) {
        vProgress = pProgress;
    }

    public void setIsAborted(boolean pIsAborted) {
        vIsAborted = pIsAborted;
    }

    public void setIsSuccessful(boolean pIsSuccessful) {
        vIsSuccessful = pIsSuccessful;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    private void toFile(HttpMethod method) {
        HttpClient client = new HttpClient();
        client.getParams().setSoTimeout(SOCKET_READ_TIMEOUT);
        try {
            int statusCode = client.executeMethod(method);
            if (statusCode == HttpStatus.SC_OK) {
                FileOutputStream output = new FileOutputStream(getFile());
                toOutputStream(method, output);

                if (getIsAborted()) {
                    if (getFile().exists()) {
                        getFile().delete();
                    }
                    LOG.error("Download of [" + getFile() + "] aborted.");
                }
            } else {
                LOG.error("Problem downloading[" + getUrl() + "]; response[" + statusCode + "].");
            }
        } catch (FileNotFoundException fnfe) {
            LOG.error("File not found.", fnfe);
        } catch (IOException ioe) {
            LOG.error("HTTP downloading problem.", ioe);
            if (getFile().exists()) {
                getFile().delete();
            }
        } finally {
            method.releaseConnection();
        }
    }


    private void toOutputStream(HttpMethod pMethod, OutputStream pOutput) throws IOException {
        try {
            int contentLength = 0;
            Header header = pMethod.getResponseHeader("Content-Length");
            if (header != null) {
                contentLength = Integer.parseInt(header.getValue());
                getProgress().setMaximum(contentLength);
            }

            byte[] buffer = new byte[BUFFER_SIZE];

            InputStream stream = pMethod.getResponseBodyAsStream();

            int bytesRead=stream.read(buffer);
            while (!getIsAborted() && (bytesRead != -1)) {
                getProgress().incrementValue(bytesRead);
                pOutput.write(buffer, 0, bytesRead);
                bytesRead=stream.read(buffer);
            }

            pOutput.close();

            boolean partialRead = (!getIsAborted() && (contentLength > 0) && (getProgress().getValue() < contentLength));
            if (partialRead) {
                LOG.error("Expected [" + contentLength + "] bytes but got only [" + getProgress().getValue() + "].");
                // Need to set the aborted flag so the file gets deleted.
                setIsAborted(true);
            }

            if (!getIsAborted()) {
                setIsSuccessful(true);
            }
        } finally {
            getProgress().finished();
        }
    }


    //------------------------------------------------------------
    //- Public Interface Functions

    public String giveStringBody() {
        String result = null;

        HttpClient client = new HttpClient();
        client.getParams().setSoTimeout(SOCKET_READ_TIMEOUT);
        GetMethod method = new GetMethod(getUrl());
        try {
            int statusCode = client.executeMethod(method);
            if (statusCode == HttpStatus.SC_OK) {
                ByteArrayOutputStream output = new ByteArrayOutputStream();
                toOutputStream(method, output);
                if (getIsSuccessful()) {
                    result = output.toString();
                }
            } else {
                LOG.error("Problem downloading[" + getUrl() + "]; response[" + statusCode + "].");
            }
        } catch (IOException ioe) {
            LOG.error("HTTP downloading problem.", ioe);
        } finally {
            method.releaseConnection();
            getProgress().finished();
        }

        return result;
    }


    public void run() {
        ThreadMonitor.update("Fired.");
        HttpClient client = new HttpClient();
        client.getParams().setSoTimeout(SOCKET_READ_TIMEOUT);
        try {
            GetMethod method = new GetMethod(getUrl());
            try {
                int statusCode = client.executeMethod(method);
                if (statusCode == HttpStatus.SC_OK) {
                    toFile(method);
                } else {
                    LOG.error("Problem downloading[" + getUrl() + "]; response[" + statusCode + "].");
                }
            } catch (IOException ioe) {
                LOG.error("HTTP downloading problem.", ioe);
            } finally {
                method.releaseConnection();
            }
        } catch (IllegalArgumentException iae) {
            LOG.error("Problem creating method.", iae);
        } finally {
            getProgress().finished();
            ThreadMonitor.update("Finished.");
        }
    }


    public void runGui(String pTitle) {
        if (GraphicsEnvironment.isHeadless()) {
            LOG.warn("GraphicsEnvironment is headless.  Running in non-GUI mode.");
        } else {
            DownloadGui gui = new DownloadGui(pTitle, this);
            gui.launch();
        }
        run();
    }


    //------------------------------------------------------------
    //- Class Interface Functions



    //------------------------------------------------------------
    //- Inner Classes



    //------------------------------------------------------------
    //- Main

    public static void main(String[] args) {
        try {
            CommandAction commands = new Command("", "<url> <filename>") {
                public boolean process(CommandArgs pArgs) {
                    boolean success = false;  // Pessimistic.
                    if (pArgs.hasNext()) {
                        String url = pArgs.next();
                        if (pArgs.hasNext()) {
                            String filename = pArgs.next();
                            if (pArgs.exhausted()) {
                                Download download = new Download(url, new File(filename));
                                download.runGui("Downloading");
                                success = true;
                            }
                        } else {
                            LOG.error("No filename given.");
                        }
                    } else {
                        LOG.error("No url given.");
                    }
                    return success;
                }
            };
            commands.process(args);
        } catch (Throwable thr) {
            LOG.error("Fatal error.", thr);
        }
    }


}
